# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A few utilities for tests."""

import sys

from absl import flags

FLAGS = flags.FLAGS


# pytest doesn't run the test as a main, so it doesn't parse the flags
# so if flags are required in tests, this will ensure that flags are manually
# parsed and the desired flag exists.
def ensure_flag(flag_str):
  try:
    getattr(FLAGS, flag_str)
  except flags.UnparsedFlagAccessError:
    # Manually parse flags.
    FLAGS(sys.argv)
  finally:
    assert getattr(FLAGS, flag_str)
